
/* demo_key.c: Sony Demo Key Driver.
 *
 * Copyright 2016 Sony Corporation.
 * Author: Sony Corporation.
 *
 * This package is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * THIS PACKAGE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/sched.h>
#include <linux/interrupt.h>
#include <linux/moduleparam.h>
#include <linux/slab.h>
#include <linux/delay.h>
#include <linux/platform_device.h>
#include <linux/miscdevice.h>
#include <linux/input.h>
#include <linux/of.h>
#include <linux/of_platform.h>
#include <linux/iio/consumer.h>
#include <linux/iio/types.h>
#include <asm/uaccess.h>
#include <asm/atomic.h>
#include <icx_key.h>
#include "./demo_key.h"

/* #define _DEBUG_DEMO_KEY_ */
#ifdef _DEBUG_DEMO_KEY_
	#define TRACE(fmt, ...)               printk("[KEY][L.%d] %s: " fmt, __LINE__, __func__, ## __VA_ARGS__)
	#define ERROR(fmt, ...)               printk("[KEY][L.%d] ### ERROR %s: " fmt, __LINE__, __func__, ## __VA_ARGS__)
#else /* _DEBUG_DEMO_KEY_ */
	#define TRACE(fmt, ...)               do {} while (0)
	#define ERROR(fmt, ...)               printk("[KEY] ### ERROR %s: " fmt, __func__, ## __VA_ARGS__)
#endif /* _DEBUG_DEMO_KEY_ */

#define DRV_NAME              "demo_key"


/* Device tree */
#define PROP_DEMO_KEY_NAME    ("demo_key")
#define PROP_KEY_DATA_NAME    ("key_data")
#define PROP_IO_CH_NAME       ("io-channel-name")

#define for_each_child_of_node_by_name(from, name, dn) \
	for (dn = of_find_node_by_name(from, name); dn; \
	     dn = of_find_node_by_name(dn, name))


/* DEMO key driver status */
enum {
	DEMO_STATUS_MODEOUT = 0,                                  /* mode out              */
	DEMO_STATUS_MODEIN,                                       /* mode in               */
	DEMO_STATUS_KEYPUSH,                                      /* key push              */
	DEMO_STATUS_AUDIO_DET,                                    /* audio-in cable detect */
};


#define POLLING_FOR_MODE_IN                   500             /* msec                  */
#define NORMAL_POLL_INTERVAL                  10              /* msec                  */
#define SCAN_INTERVAL_CHAT_GUARD              10              /* msec                  */
#define SCAN_INTERVAL_CHAT_GUARD_CABLE_DET    500             /* msec, for cable det   */
#define ADIN2_ERROR_TRIAL_WAIT                50              /* msec                  */
#define AUDIOIN_CABLE_MONITOR_INTERVAL        500             /* msec                  */
#define CHAT_TRIAL_CNT                        3
#define ADIN2_ERROR_TRIAL_CNT                 3
#define DEMO_KEY_MAX                          10              /* Dummy                 */


/* AD key info */
struct dmkey_akeyinfo {
	const char                  *channel;                     /* Channel name          */
	int                          count;                       /* AD key data count     */
	struct dmkey_akeydata        data[DEMO_KEY_MAX];          /* AD key data           */
};

/* Private data */
struct priv {
	struct miscdevice            misc;                        /* MISC device           */
	struct input_dev            *p_input;                     /* Input device          */
	struct delayed_work          dwork;                       /* Delay work            */
	u8                           key_count;                   /* Key count             */
	u8                           dummy[3];                    /* Dummy                 */
	u32                          p_key_code[DEMO_KEY_MAX];    /* Key code list         */
	struct iio_channel          *chan;                        /* AD channel            */
};


static u8                        demo_status;
static int                       demo_key_code     = KEY_RESERVED;
static int                       pre_demo_key_code = KEY_RESERVED;
static struct priv              *m_p_demo_key;
static struct dmkey_akeyinfo     ad_2_code_table;


/* -------------------------------------------------------------------------------- */
/*   AD key info setting                                                            */
/*                                                                                  */
/*   IN : p_priv = private data                                                     */
/*        p_node = device node                                                      */
/*   OUT: 0                                                                         */
/*        -1    == Error                                                            */
/* -------------------------------------------------------------------------------- */
static int set_akey_info(struct priv *p_priv, struct device_node *p_node)
{
	uint        count        = 0;
	int         keydata_size = 0;
	int         result       = 0;

	struct device_node          *p_child = NULL;
	struct dmkey_akeydata const *p_data  = NULL;

	for_each_child_of_node_by_name(p_node, PROP_DEMO_KEY_NAME, p_child) {
		p_data = of_get_property(p_child, PROP_KEY_DATA_NAME, &keydata_size);
		if (p_data) {
			of_property_read_u32_array(p_child, PROP_KEY_DATA_NAME, (u32 *)&ad_2_code_table.data[0], (keydata_size / sizeof(int)));
			result = of_property_read_string(p_child, PROP_IO_CH_NAME, &(ad_2_code_table.channel));
			if (!result) {
				ad_2_code_table.count = keydata_size / sizeof(struct dmkey_akeydata);
				if (ad_2_code_table.count > DEMO_KEY_MAX) {
					ERROR("%s(%d) : ad_2_code_table.count[%d]\n", __func__, __LINE__, ad_2_code_table.count);
					return -1;
				}

				for (count = 0; count < ad_2_code_table.count; count++) {
					p_priv->p_key_code[p_priv->key_count++] = ad_2_code_table.data[count].key_code;
					set_bit(ad_2_code_table.data[count].key_code, p_priv->p_input->keybit);
				}
			} else {
				ERROR("%s(%d) : result[%d]\n", __func__, __LINE__, result);
			}
		} else {
			ERROR("%s(%d) : p_data[%s]\n", __func__, __LINE__, (p_data)?"Not NULL":"NULL");
		}
	}

	return 0;
}


/* -------------------------------------------------------------------------------- */
/*   DEMO key sleep                                                                 */
/*                                                                                  */
/*   IN : timeout = sleep time out                                                  */
/*   OUT: none                                                                      */
/* -------------------------------------------------------------------------------- */
static void key_sleep(int timeout)
{
	set_current_state(TASK_INTERRUPTIBLE);
	schedule_timeout(timeout);
}


/* -------------------------------------------------------------------------------- */
/*   DEMO key down                                                                  */
/*                                                                                  */
/*   IN : key_code = key code                                                       */
/*   OUT: none                                                                      */
/* -------------------------------------------------------------------------------- */
static void input_sync_key_down(int key_code)
{
	input_event(m_p_demo_key->p_input, EV_KEY, key_code, DEMO_KEY_VALUE_DOWN);
	input_sync(m_p_demo_key->p_input);
}


/* -------------------------------------------------------------------------------- */
/*   DEMO key up                                                                    */
/*                                                                                  */
/*   IN : key_code = key code                                                       */
/*   OUT: none                                                                      */
/* -------------------------------------------------------------------------------- */
static void input_sync_key_up(int key_code)
{
	input_event(m_p_demo_key->p_input, EV_KEY, key_code, DEMO_KEY_VALUE_UP);
	input_sync(m_p_demo_key->p_input);
}


/* -------------------------------------------------------------------------------- */
/*   Get key code from AD value                                                     */
/*                                                                                  */
/*   IN : ad_val = AD value                                                         */
/*   OUT: key code                                                                  */
/*        -1 == Error                                                               */
/* -------------------------------------------------------------------------------- */
static inline int key_code_from_ad_val(int ad_val)
{
	int i;

	for (i = 0; i < ad_2_code_table.count; i++)
		if ((ad_2_code_table.data[i].min <= ad_val) &&
			(ad_val <= ad_2_code_table.data[i].max)) {
			return ad_2_code_table.data[i].key_code;
		}

	return -1;
}


/* -------------------------------------------------------------------------------- */
/*   Get AD value                                                                   */
/*                                                                                  */
/*   IN : chan = AD channel                                                         */
/*   OUT: AD value                                                                  */
/*        -1 == Error                                                               */
/* -------------------------------------------------------------------------------- */
static int get_md_adc_val(struct iio_channel *chan)
{
	int val           = 0;
	int ret           = 0;
	int key_code_pre  = KEY_RESERVED;
	int key_code_cur  = KEY_RESERVED;
	int chat_cnt      = 0;
	int chat_interval = 0;

	if (demo_status == DEMO_STATUS_MODEOUT) {
		chat_interval = SCAN_INTERVAL_CHAT_GUARD_CABLE_DET;
	} else {
		chat_interval = SCAN_INTERVAL_CHAT_GUARD;
	}

	while (chat_cnt < CHAT_TRIAL_CNT) {
		if (chat_cnt) {
			key_sleep(msecs_to_jiffies(chat_interval));
		}
		ret = iio_read_channel_processed(chan, &val);
		/* TRACE("adc val=%d\n", val>>4); */
		if (ret >= 0) {
			key_code_cur = key_code_from_ad_val(val);
			if ((key_code_pre == KEY_RESERVED) || (key_code_pre == key_code_cur)) {
				key_code_pre = key_code_cur;
				chat_cnt++;
			} else {
				key_code_pre = KEY_RESERVED;
				key_code_cur = KEY_RESERVED;
				chat_cnt     = 0;
			}
		} else {
			return -1;
		}
	}

	return val;
}


/* -------------------------------------------------------------------------------- */
/*   DEMO key work queue process                                                    */
/*                                                                                  */
/*   IN : work                                                                      */
/*   OUT: none                                                                      */
/* -------------------------------------------------------------------------------- */
static void workq_func_demo(struct work_struct *work)
{
	int ad_val;
	int i = 0;

	/* reading ADIN2 value */
	for (i = 0; i < ADIN2_ERROR_TRIAL_CNT; i++) {
		ad_val = get_md_adc_val(m_p_demo_key->chan);
		if (!(ad_val < 0)) {
			break;
		}
		key_sleep(msecs_to_jiffies(ADIN2_ERROR_TRIAL_WAIT));
	}
	if (ad_val < 0) {
		ERROR("get_md_adc_val() failed\n");
		return;
	}

	demo_key_code = key_code_from_ad_val(ad_val);

	switch (demo_status) {
	case DEMO_STATUS_MODEOUT:
		if (demo_key_code == ad_2_code_table.data[0].key_code) {
			/* demo jig plugged */
			input_sync_key_down(ad_2_code_table.data[0].key_code);
			demo_status = DEMO_STATUS_MODEIN;
			schedule_delayed_work(&m_p_demo_key->dwork, msecs_to_jiffies(NORMAL_POLL_INTERVAL));
			TRACE("[Demo Driver] Demo In\n");
		} else if (demo_key_code == ad_2_code_table.data[1].key_code) {
			/* audio-in cable plugged */
			input_sync_key_down(ad_2_code_table.data[1].key_code);
			demo_status = DEMO_STATUS_AUDIO_DET;
			schedule_delayed_work(&m_p_demo_key->dwork, msecs_to_jiffies(AUDIOIN_CABLE_MONITOR_INTERVAL));
			TRACE("[Demo Driver] Audio cable plugged\n");
		} else {
			schedule_delayed_work(&m_p_demo_key->dwork, msecs_to_jiffies(POLLING_FOR_MODE_IN));
		}
		break;
	case DEMO_STATUS_MODEIN:
		if (ad_val > ad_2_code_table.data[0].max) {
			/* demo jig unplugged */
			input_sync_key_up(ad_2_code_table.data[0].key_code);
			demo_status = DEMO_STATUS_MODEOUT;
			schedule_delayed_work(&m_p_demo_key->dwork, msecs_to_jiffies(POLLING_FOR_MODE_IN));
			TRACE("[Demo Driver] Demo Out\n");
		} else {
			if (pre_demo_key_code != demo_key_code) {
				pre_demo_key_code = demo_key_code;
				if (demo_key_code != ad_2_code_table.data[0].key_code) {
					/* demo key pushed */
					input_sync_key_down(pre_demo_key_code);
					demo_status = DEMO_STATUS_KEYPUSH;
					TRACE("key code=%d\n", pre_demo_key_code);
					TRACE("key pressed\n");
				}
			}
			schedule_delayed_work(&m_p_demo_key->dwork, msecs_to_jiffies(NORMAL_POLL_INTERVAL));
		}
		break;
	case DEMO_STATUS_KEYPUSH:
		/* polling for demo key release */
		input_sync_key_up(pre_demo_key_code);
		demo_status = DEMO_STATUS_MODEIN;
		TRACE("key code=%d\n", pre_demo_key_code);
		TRACE("key released\n");

		if (ad_val > ad_2_code_table.data[0].max) {
			/* demo jig unplugged */
			input_sync_key_up(ad_2_code_table.data[0].key_code);
			demo_status = DEMO_STATUS_MODEOUT;
			schedule_delayed_work(&m_p_demo_key->dwork, msecs_to_jiffies(POLLING_FOR_MODE_IN));
			TRACE("[Demo Driver] Demo Out\n");
		} else {
			schedule_delayed_work(&m_p_demo_key->dwork, msecs_to_jiffies(NORMAL_POLL_INTERVAL));
		}
		break;
	case DEMO_STATUS_AUDIO_DET:
		if (ad_val > ad_2_code_table.data[1].max) {
			/* audio-in cable unplugged */
			input_sync_key_up(ad_2_code_table.data[1].key_code);
			demo_status = DEMO_STATUS_MODEOUT;
			schedule_delayed_work(&m_p_demo_key->dwork, msecs_to_jiffies(POLLING_FOR_MODE_IN));
			TRACE("[Demo Driver] Audio cable unplugged\n");
		} else {
			schedule_delayed_work(&m_p_demo_key->dwork, msecs_to_jiffies(AUDIOIN_CABLE_MONITOR_INTERVAL));
		}
		break;
	default:
		/* nop */
		break;
	}
}


/* -------------------------------------------------------------------------------- */
/*   Open                                                                           */
/*                                                                                  */
/*   IN : p_inode = i node                                                          */
/*        p_file  = file                                                            */
/*   OUT: 0                                                                         */
/* -------------------------------------------------------------------------------- */
static int dm_key_open(struct inode *p_inode, struct file *p_file)
{
	p_file->private_data = container_of(p_file->private_data, struct priv, misc) ;

	return 0 ;
}


/* -------------------------------------------------------------------------------- */
/*   ioctl                                                                          */
/*                                                                                  */
/*   IN : p_file = file                                                             */
/*        cmd    = command                                                          */
/*        arg    = argument                                                         */
/*   OUT: 0                                                                         */
/*        key count                                                                 */
/*        -EFAULT                                                                   */
/*        -EINVAL                                                                   */
/* -------------------------------------------------------------------------------- */
static long dm_key_ioctl(struct file *p_file, unsigned int cmd, unsigned long arg)
{
	struct priv *p_demo_key = p_file->private_data;

	switch (cmd) {
	case ICX_KEY_IOCTL_CMD_GETCOUNT:
	{
		return p_demo_key->key_count ;
	}

	case ICX_KEY_IOCTL_CMD_GETINFO:
	{
		if (copy_to_user((void __user *)arg, &p_demo_key->p_key_code[0], sizeof(u32)*p_demo_key->key_count)) {
			return -EFAULT ;
		}
		break ;
	}

	default:
		return -EINVAL ;
	}

	return 0 ;
}


static const struct file_operations		dm_key_fops = {
											.owner          = THIS_MODULE,
											.open           = dm_key_open,
											.unlocked_ioctl = dm_key_ioctl,
											.compat_ioctl   = dm_key_ioctl,
										} ;


/* -------------------------------------------------------------------------------- */
/*   Suspend                                                                        */
/*                                                                                  */
/*   IN : p_dev = platform device                                                   */
/*        state = status                                                            */
/*   OUT: 0                                                                         */
/* -------------------------------------------------------------------------------- */
static int dm_key_suspend(struct platform_device *p_dev, pm_message_t state)
{
	struct priv *p_demo_key = platform_get_drvdata(p_dev);

	TRACE("entering suspend\n");
	cancel_delayed_work(&p_demo_key->dwork);

	/* demo jig unplugged */
	input_sync_key_up(ad_2_code_table.data[0].key_code);
	demo_status = DEMO_STATUS_MODEOUT;
	schedule_delayed_work(&m_p_demo_key->dwork, msecs_to_jiffies(POLLING_FOR_MODE_IN));
	TRACE("[Demo Driver] Demo Out\n");

	return 0;
}


/* -------------------------------------------------------------------------------- */
/*   Resume                                                                         */
/*                                                                                  */
/*   IN : p_dev = platform device                                                   */
/*   OUT: 0                                                                         */
/* -------------------------------------------------------------------------------- */
static int dm_key_resume(struct platform_device *p_dev)
{
	struct priv *p_demo_key = platform_get_drvdata(p_dev);

	TRACE("entering resume\n");
	schedule_delayed_work(&p_demo_key->dwork, 0);

	return 0;
}


/* -------------------------------------------------------------------------------- */
/*   Remove                                                                         */
/*                                                                                  */
/*   IN : p_dev = platform device                                                   */
/*   OUT: 0                                                                         */
/* -------------------------------------------------------------------------------- */
static int dm_key_remove(struct platform_device *p_dev)
{
	struct priv *p_demo_key = platform_get_drvdata(p_dev);
	TRACE("entering remove\n");

	/* Free ADC */
	iio_channel_release(p_demo_key->chan);

	cancel_delayed_work_sync(&p_demo_key->dwork);
	input_unregister_device(p_demo_key->p_input);
	misc_deregister(&p_demo_key->misc);
	kfree(p_demo_key);

	return 0;
}


/* -------------------------------------------------------------------------------- */
/*   Probe                                                                          */
/*                                                                                  */
/*   IN : p_dev = platform device                                                   */
/*   OUT: 0                                                                         */
/*        -ENOMEM                                                                   */
/* -------------------------------------------------------------------------------- */
static int dm_key_probe(struct platform_device *p_dev)
{
	int          result;
	struct priv *p_demo_key;
	uint         size;
	uint         i;

	demo_status = DEMO_STATUS_MODEOUT;

	printk(KERN_INFO "Starting Demo Mode Key Driver.\n");

	size = sizeof(struct priv);
	p_demo_key = kzalloc(size, GFP_KERNEL);
	if (p_demo_key == NULL) {
		ERROR("%s(%d)\n", __func__, __LINE__);
		return -ENOMEM;
	}

	INIT_DELAYED_WORK(&p_demo_key->dwork, workq_func_demo);

	/* MISC device register */
	p_demo_key->misc.minor = MISC_DYNAMIC_MINOR ;
	p_demo_key->misc.name  = DRV_NAME ;
	p_demo_key->misc.fops  = &dm_key_fops ;
	result = misc_register(&p_demo_key->misc) ;
	if (result) {
		ERROR("%s(%d)\n", __func__, __LINE__);
		goto dm_key_probe_f00;
	}

	/* allocate input_dev and initialize it. */
	p_demo_key->p_input = input_allocate_device();
	if (p_demo_key->p_input == NULL) {
		result = -ENOMEM ;
		ERROR("%s(%d)\n", __func__, __LINE__);
		goto dm_key_probe_e00;
	}

	set_bit(EV_KEY, p_demo_key->p_input->evbit) ;
	p_demo_key->p_input->name       = DRV_NAME ;
	p_demo_key->p_input->phys       = DRV_NAME ;
	p_demo_key->p_input->id.bustype = BUS_HOST ;
	p_demo_key->p_input->id.vendor  = 0x0001 ;
	p_demo_key->p_input->id.product = 0x0001 ;
	p_demo_key->p_input->id.version = 0x0100 ;
	p_demo_key->p_input->dev.parent = &p_dev->dev ;

	/* Set AD key */
	result = set_akey_info(p_demo_key, p_dev->dev.of_node);
	if (result) {
		ERROR("%s(%d)\n", __func__, __LINE__);
		goto dm_key_probe_d00;
	}

	for (i = 0; i < REP_CNT; i++) {
		p_demo_key->p_input->rep[i] = 0;
	}

	result = input_register_device(p_demo_key->p_input);
	if (result != 0) {
		ERROR("%s(%d)  result=%d\n", __func__, __LINE__, result);
		goto dm_key_probe_d00;
	}

	/* Get ADC channel */
	p_demo_key->chan = iio_channel_get(&p_dev->dev, ad_2_code_table.channel);

	platform_set_drvdata(p_dev, p_demo_key) ;
	m_p_demo_key = p_demo_key;

	/* going to scan ad_key right away. */
	schedule_delayed_work(&p_demo_key->dwork, 0);

	return 0 ;

dm_key_probe_d00:
	input_free_device(p_demo_key->p_input);
dm_key_probe_e00:
	misc_deregister(&p_demo_key->misc);
dm_key_probe_f00:
	kfree(p_demo_key);

	return result;
}


static const struct of_device_id demo_key_of_match[] = {
	{ .compatible = "sony,demo-key" },
	{ },
};

MODULE_DEVICE_TABLE(of, demo_key_of_match);


static struct platform_driver		dm_key_platform_driver = {
										.probe     = dm_key_probe,
										.remove    = dm_key_remove,
										.suspend   = dm_key_suspend,
										.resume    = dm_key_resume,
										.driver    = {
											.name  = DRV_NAME,
											.owner = THIS_MODULE,
											.of_match_table = demo_key_of_match,
										},
									};


module_platform_driver(dm_key_platform_driver);
MODULE_DESCRIPTION("Sony Demo Key Driver");
MODULE_AUTHOR("Sony Corporation");
MODULE_LICENSE("GPL");
MODULE_ALIAS("Sony Demo Key Driver");
MODULE_VERSION("v1.00");

